package com.jdbernard.wdiwtlt.cli

import com.jdbernard.wdiwtlt.ConfigWrapper
import com.jdbernard.wdiwtlt.MediaLibrary
import com.jdbernard.wdiwtlt.db.ORM
import com.jdbernard.wdiwtlt.db.models.*
import com.jdbernard.io.NonBlockingInputStreamReader
import com.jdbernard.util.AnsiEscapeCodeSequence as ANSI
import com.zaxxer.hikari.HikariConfig
import com.zaxxer.hikari.HikariDataSource
import org.docopt.Docopt
import jline.console.ConsoleReader
import java.util.regex.Matcher
import org.slf4j.Logger
import org.slf4j.LoggerFactory

import static com.jdbernard.util.AnsiEscapeCodeSequence.*

public class CommandLineInterface {

    public static final VERSION = "ALPHA"

    public static final def DOC = """\
wdiwtlt v$VERSION

Usage:
    wdiwtlt [options]
    wdiwtlt --version
    wdiwtlt --help

Options:
    -c --config <config-file>

        Config file for wdiwtlt CLI.

    -L --library-root <root-dir>

        The path to a local media library directory. Providing a local library
        causes wdiwtlt to run in local mode.

    -D --database-config <config-file>

        The path to a data source configuration file. wdiwtlt cli uses HikariCP
        for database connection pooling, so the given configuration file is
        expected to be formatted so that it can be fed directly to the
        HikariConfig constructor.
        (see https://github.com/brettwooldridge/HikariCP#configuration-knobs-baby)

Configuration:
"""

    private static Logger logger =
        LoggerFactory.getLogger(CommandLineInterface)
    private Properties cliConfig
    private MediaLibrary library

    private InputStream inStream
    private OutputStream outStream

    private ConsoleReader reader
    private Thread consoleReaderThread
    private List<String> consoleReadBuffer =
        Collections.synchronizedList(new ArrayList<String>())
    private synchronized boolean running

    private String titleStyle, normalStyle, statusStyle, promptStyle,
        artistStyle, albumStyle, fileStyle, errorStyle, playlistStyle
    private String clearLine = new ANSI().eraseLine(Erase.All).toString()
    private String afterInput =
        new ANSI().eraseLine(Erase.All).scrollUp().cursorUp().toString()
    private String beforeLeader =
        new ANSI().saveCursor().cursorPrevLine(2).toString()
    private String afterLeader =
        new ANSI().restoreCursor().eraseLine(Erase.ToEnd).toString()
    private String eraseLeader =
        new ANSI().cursorPrevLine().eraseLine(Erase.All).cursorPrevLine()
            .eraseLine(Erase.All).toString()

    private int displayWidth = 79
    private long msgTimeout
    private ScrollText currentlyPlaying = new ScrollText(
        maxWidth: displayWidth - 15 - VERSION.size(),
        text: "No media currently playing.")
    private ScrollText status = new ScrollText(maxWidth: displayWidth)
    private Date dismissMsgDate = new Date()

    def selection = [:]

    public static void main(String[] args) {

        def opts = new Docopt(DOC).withVersion("wdiwtlt v$VERSION").parse(args)

        def exitErr = { msg ->
            System.err.println("wdiwtlt: $msg")
            System.exit(1) }

        // Look for a given CLI config file.
        def givenCfg = new Properties()
        File cfgFile
        if (opts["--config"]) cfgFile = new File(opts["--config"])

        if (!cfgFile || !cfgFile.exists() || !cfgFile.isFile())
            cfgFile = new File("wdiwtlt.cli.properties")

        if (!cfgFile || !cfgFile.exists() || !cfgFile.isFile())
            cfgFile = new File(System.getenv().HOME, ".wdiwtlt.cli.properties")

        if (cfgFile.exists() && cfgFile.isFile()) {
            try { cfgFile.withInputStream { givenCfg.load(it) } }
            catch (Exception e) { givenCfg.clear() } }

        // Just in case, load up the general WDIWTLT default config service
        def wdiwtltDefaultConfig = new ConfigWrapper()

        // Find the library root directory (TODO: eventually this should be
        // enabled only when local mode is chosen).
        File libRoot = new File(
            opts["--library-root"] ?:
            givenCfg[ConfigWrapper.LIBRARY_DIR_KEY] ?:
            wdiwtltDefaultConfig.libraryRootPath)

        if (libRoot && (!libRoot.exists() || !libRoot.isDirectory()))
            exitErr("Library root does not exist or is not a directory: " +
                libRoot.canonicalPath)

        // Get the library database
        HikariConfig hcfg
        File dbCfgFile = new File(
            opts["--database-config"] ?:
            givenCfg[ConfigWrapper.DB_CONFIG_FILE_KEY] ?:
            "database.properties")

        // Try to load from the given DB config file
        if (dbCfgFile && dbCfgFile.exists() && dbCfgFile.isFile()) {
            Properties props = new Properties()
            try { dbCfgFile.withInputStream { props.load(it) } }
            catch (Exception e) { props.clear() }

            if (props) hcfg = new HikariConfig(props) }

        // Look to see if database connection properties are given in the CLI
        // config file.
        if (givenCfg && givenCfg["database.config.dataSourceClassName"]) {
            Properties props = new Properties()
            props.putAll(givenCfg
                .findAll { it.key.startsWith("database.config.") }
                .collectEntries { [it.key[16..-1], it.value] } )
            if (props) hcfg = new HikariConfig(props) }

        // Fall back to the WDIWTLT general database defaults
        if (!hcfg) hcfg = wdiwtltDefaultConfig.hikariConfig
        if (!hcfg) exitErr("Cannot load database configuation.")

        // Create our database instance
        def orm
        try {
            def hds = new HikariDataSource(hcfg)
            orm = new ORM(hds) }
        catch (Exception e) {
            exitErr("Could not establish a connection to the database:\n\t" +
                e.localizedMessage) }

        // Create our CLI object
        def cliInst = new CommandLineInterface(new MediaLibrary(orm, libRoot),
            System.in, System.out, givenCfg)
        cliInst.repl()
    }

    public CommandLineInterface(MediaLibrary library, InputStream sin,
    OutputStream out, Properties cliConfig) {
        this.cliConfig = cliConfig
        this.library = library
        this.inStream = sin
        this.outStream = out

        this.msgTimeout = cliConfig["message.timeout"] ?: 5000l

        setupTextStyles()

        this.consoleReaderThread = new Thread().start(this.&runReaderThread)
    }

    void setupTextStyles() {
        titleStyle = new ANSI().color(Colors.GREEN, Colors.DEFAULT, false).toString()
        normalStyle = new ANSI().resetText().toString()
        promptStyle = new ANSI().color(Colors.YELLOW, Colors.DEFAULT, true).toString()
        statusStyle = new ANSI().color(Colors.CYAN, Colors.DEFAULT, false).toString()
        artistStyle = new ANSI().color(Colors.RED, Colors.DEFAULT, false).toString()
        albumStyle = new ANSI().color(Colors.BLUE, Colors.DEFAULT, false).toString()
        fileStyle = new ANSI().color(Colors.GREEN, Colors.DEFAULT, false).toString()
        errorStyle = new ANSI().color(Colors.RED, Colors.DEFAULT, true).toString()
    }

    private void runReaderThread() {
        this.reader = new ConsoleReader(this.inStream, this.outStream)
        this.reader.setPrompt(promptStyle + "> " + normalStyle)

        String line
        while(this.running && !Thread.currentThread().isInterrupted()) {
            line = reader.readLine()
            if (line == null || line == "quit" || line == "exit") running = false
            else {
                consoleReadBuffer.add(line)
                outStream.print(afterInput)
                outStream.flush() } } }

    public void repl() {
        this.running = true
        String line = null

        outStream.println ""
        outStream.println ""
        drawLeader()

        while(this.running) {

            if (new Date() > dismissMsgDate) resetStatus()
            if (consoleReadBuffer.size() > 0) {
                line = consoleReadBuffer.remove(0)
                processInput(line.split(/\w/) as LinkedList)
            } else {
                drawLeader()
                Thread.sleep(250)
            }
        }
    }

    private def processInput(LinkedList<String> line) {
        String command = line.poll()
        switch(command.toLowerCase()) {
            case 'album':   return selectAlbum(line)
            case 'artist':  return selectArtist(line)
            case 'scan':    return scanMediaLibrary()

            case 'list':
                String nextArg = line.poll()
                if (nextArg.toLowerCase() == 'all')
                    return processList(line, true)
                else {
                    if (nextArg) line.push(nextArg)
                    return processList(line, false) }

            case 'debug':
                outStream.println(
                    "\n\nConfig: \n" +
                    cliConfig.collect { "\t${it.key}: ${it.value}" }
                             .join("\n") +
                    "\n\n\n")
                drawLeader(true)
                return

            case 'q': case 'quit': case ':q': case 'exit': case '\u0004':
                running = false
                consoleReaderThread.interrupt()
                return

                selection.album = null
                resetStatus()
                break

            case "list artists for album":
                if (!selection.album) {
                    setErr("No album is selected.")
                    break }
                outStream.println(makeArtistList(
                    library.getArtistsByAlbumId(selection.album.id),
                    selection.album))
                drawLeader(true)
                break

            case ~/list artists( .+)?/:
                def name = Matcher.lastMatcher[0][1]?.trim()
                def artists
                if (name) artists = library.getArtistsByName(name)
                else artists = library.getArtists()

                outStream.println(makeArtistList(artists))
                drawLeader(true)
                break

            case "list albums for artist":
                if (!selection.artist) {
                    setErr("No artist selected.")
                    break }
                outStream.println(makeAlbumList(
                    library.getAlbumsByArtistId(selection.artist.id),
                    selection.artist))
                drawLeader(true)
                break

            case ~/list albums( .+)?/:
                def name = Matcher.lastMatcher[0][1]?.trim()
                def albums
                if (name) albums = library.getAlbumsByName(name)
                else albums = library.getAlbums()

                outStream.println(makeAlbumList(albums))
                drawLeader(true)
                break

            default:
                status.text = errorStyle +
                    "Unrecognized command: '$line'${normalStyle}"
                dismissMsgDate = new Date(new Date().time + msgTimeout)
                drawLeader()
                Thread.sleep(250)
                break
        }

    }

    private def processList(LinkedList options, boolean all) {
        def option = options.poll()
        switch(option) {
            case 'albums':
            case 'artists':
            case 'bookmarks':
            case 'playlists':
            default:
                printLongMessage("Unrecognized option to the ${promptStyle}" +
                    "list${normalStyle} command. Use ${promptStyle}help list" +
                    "to see a list of valid options.")
                return null
        }
    }

    public Album selectAlbum(LinkedList input) {
        String criteria = input.join(" ")

        if (!criteria) { selection.album = null; resetStatus(); return null }

        Album match = library.getByIdOrName(Album, criteria)

        if (!match) { setErr("No album matches '$input'."); return null }
        else if (match.size() > 1) {
            setErr("Multiple albums match '$input': " +
                match.collect { "${it.id}: ${it.name}" }.join(", "))
            return null }

        selection.album = match[0]
        resetStatus()
        return match[0] }

    public void selectArtist(LinkedList input) {
        String criteria = input.join(" ")

        if (!criteria) { selection.artist = null; resetStatus(); return null }

        Artist match = library.getByIdOrName(Artist, criteria)

        if (!match) { setErr("No artist matches '$input'."); return null }
        else if (match.size() > 1) {
            setErr("Multiple artists match '$input': " +
                match.collect { "${it.id}: ${it.name}" }.join(", "))
            return null }
        selection.artist = match[0]
        resetStatus()
        return match[0] }

    public MediaLibrary scanMediaLibrary() {
        status.text = "Scanning media library..."
        library.rescanLibrary()
        status.text = "Scanned ? files."
        dismissMsgDate = new Date(new Date().time + msgTimeout)
        return library }

    private void drawLeader(afterOutput = false) {

        String leader = beforeLeader + getLeader() +
            (afterOutput ?
                (promptStyle + "> " + normalStyle) :
                afterLeader)

        outStream.print(leader)
        outStream.flush() }

    private void printLongMessage(String msg) {
        String result = new StringBuilder()
            .append(eraseLeader)
            .append(msg)
            .append("\n\n")
            .toString()

        outStream.println result
        drawLeader(true) }

    private String getLeader() {
        StringBuilder leader = new StringBuilder()
            .append(clearLine)
            .append(titleStyle)
            .append("WDIWTLT - v")
            .append(VERSION)
            .append(normalStyle)
            .append(" -- ")
            .append(statusStyle)
            .append(currentlyPlaying)
            .append("\n")
            .append(clearLine)
            .append(normalStyle)
            .append(status)
            .append("\n")

        return leader.toString() }

    private String setErr(String errMsg) {
        status.text = errorStyle + errMsg
        dismissMsgDate = new Date(new Date().time + msgTimeout) }

    private String makeAlbumList(def albums, Artist artist = null) {
        def result = new StringBuilder()
            .append(eraseLeader)
            .append("--------------------\nAlbums")

        if (artist) result.append(" (for artist '$artist'):\n")
        else result.append(":\n\n")

        result.append(albums.collect { "${it.id}: ${it}" }.join("\n"))
            .append("\n\n\n")

        return result.toString() }

    private String makeArtistList(def artists, Album album = null) {
        def result = new StringBuilder()
            .append(eraseLeader)
            .append("--------------------\nArists")

        if (album) result.append(" (for album '$album'):\n")
        else result.append(":\n\n")

        result.append(artists.collect { "${it.id}: ${it.name}" }.join("\n"))
            .append("\n\n\n")

        return result.toString() }

    private String resetStatus() {
        StringBuilder s = new StringBuilder()

        if (selection.playlist) s.append(playlistStyle)
            .append(selection.playlist)
            .append(normalStyle)
            .append(": ")

        if (selection.artist) s.append(artistStyle)
            .append(selection.artist)
            .append(normalStyle)
            .append(" / ")

        if (selection.album) s.append(albumStyle)
            .append(selection.album)
            .append(normalStyle)
            .append(" / ")

        if (selection.mediaFile) s.append(fileStyle)
            .append(selection.mediaFile)

        if (s.size() == 0) status.text = "No current media selections."
        else status.text = s.toString()

        return status.text}
}
