package com.jdbernard.wdiwtlt

import com.jdbernard.wdiwtlt.db.ORM
import com.jdbernard.wdiwtlt.db.models.*

import org.jaudiotagger.audio.AudioFile
import org.jaudiotagger.audio.AudioFileIO
import org.jaudiotagger.tag.Tag as JATag
import org.jaudiotagger.tag.FieldKey
import org.slf4j.Logger
import org.slf4j.LoggerFactory
import org.apache.commons.codec.digest.DigestUtils

import static org.jaudiotagger.tag.FieldKey.*

public class MediaLibrary {

    private static Logger logger = LoggerFactory.getLogger(MediaLibrary)

    @Delegate ORM orm
    private File libraryRoot

    public MediaLibrary(ORM orm, File rootDir) {
        logger.debug("Creating a MediaLibrary rooted at: {}",
            rootDir.canonicalPath)

        this.orm = orm
        this.libraryRoot = rootDir }

    public void clean() {
        orm.removeEmptyAlbums()
        orm.removeEmptyArtists()
        orm.removeEmptyPlaylists()
    }

    public void rescanLibrary() { libraryRoot.eachFileRecurse { addFile(it) } }

    public MediaFile addFile(File f) {
        if (!f.exists() || !f.isFile()) {
            logger.info("Ignoring non-existant file: {}", f.canonicalPath)
            return null }

        def relPath = getRelativePath(libraryRoot, f)
        MediaFile mf = orm.getMediaFileByFilePath(relPath)

        if (mf) {
            logger.info(
                "Ignoring a media file I already know about: {}", relPath)
            return mf }

        // Read in the media's tags
        mf = new MediaFile()
        def af

        try { af = AudioFileIO.read(f) }
        catch (Exception e) {
            logger.info("Ignoring a file because I can't" +
                "read the media tag info:\n\t{}",
                e.localizedMessage)
            return null }

        def fileTag = af.tag

        mf.name = fileTag?.getFirst(TITLE)?.trim() ?: f.name
        mf.filePath = relPath
        mf.comment = fileTag?.getAll(COMMENT)?.collect { it.trim() }.join('\n\n')
        mf.trackNumber = safeToInteger(fileTag?.getFirst(TRACK))

        def folderParts = mf.filePath.split("[\\\\/]")[1..<-1] as LinkedList

        // Find artist and album names (if any)
        def artistNames = fileTag?.getAll(ARTIST)?.collect { it.trim() }
        def albumNames = fileTag?.getAll(ALBUM)?.collect { it.trim() }

        if (!artistNames) {
            mf.metaInfoSource = MediaFile.FILE_LOCATION
            artistNames = folderParts.size() >= 2 ?  [folderParts[0]] : [] }

        if (!albumNames) {
            mf.metaInfoSource = MediaFile.FILE_LOCATION
            albumNames = [folderParts.peekLast()] }

        // Hash the file
        mf.fileHash = f.withInputStream { DigestUtils.md5Hex(it) }

        orm.withTransaction {
            orm.create(mf)
            associateWithArtistsAndAlbums(mf, artistNames, albumNames, fileTag)
        }
    }

    public def getByIdOrName(Class modelClass, String input) {
        def match
        if (safeToInteger(input))  match = [orm.getById(modelClass, safeToInteger(input))]
        else {
            match = orm.getByName(modelClass, input)
            if (!match) match = orm.getLike(modelClass, ["name"], [input]) }
        return match }

    public List<Artist> getArtistsByName(String name) {
        return [orm.getArtistByName(name)] ?: orm.getArtistsLikeName(name) }

    public List<Album> getAlbumsByName(String name) {
        return orm.getAlbumsByName(name) ?: orm.getAlbumsLikeName(name) }

    private void associateWithArtistsAndAlbums(MediaFile mf, String artistNames,
    String albumNames, Integer albumYear) {

        def albums = []

        // Find or create artists.
        def artists = artistNames.collect { artistName ->
            Artist artist = orm.getArtistByName(artistName)
            if (!artist) artist = orm.create(new Artist(name: artistName))
            return artist }

        // Associate file with artists.
        artists.each { orm.associate(artist, mf) }

        // Find or create albums
        def albums = albumNames.collect { albumName ->
            Album album

            // If we know what year the album was released we can use that to
            // narrow down the list of matching albums.
            if (albumYear != null) {
                // Look first to see if we already know about this album
                // associated with one of the artists for this piece.
                album = artists.inject(null, { foundAlbum, artist ->
                    if (foundAlbum) return foundAlbum
                    return orm.getAlbumByNameAndYearAndArtistId(albumName,
                        albumYear, artist.id) }
                // If we don't have it with one of the artists, see if we have
                // one that matches the name and year.
                if (!album)
                    album = orm.getAlbumByNameAndYear(albumName, albumYear) }

            else {
                album = artists.inject(null, { foundAlbum, artist ->
                    if (foundAlbum) return foundAlbum
                    return orm.getAlbumByNameAndArtistId(albumName, artist.id) }
                if (!album) {
                    def foundAlbums = org.getAlbumsByName(albumName)
                    album = foundAlbums ? foundAlbums[0] : 0 } }

            // We still can't find the album at all. We'll need to create it
            if (!album)
                album = orm.create(new Album(name: albumName, albumYear))

            return album }

        // Associate file with albums
        albums.each { orm.associate(album, mf) }

        // Make sure we have association between all of the artists and albums.
        artists.each { artist ->
            def albumsForArtist = orm.getAlbumsByArtistId(artist.id)
            def albumsMissing = albums - albumsForArtist
            albumsMissing.each { album -> orm.associate(artist, album) } }
    }

    /** #### `getRelativePath`
      * Given a parent path and a child path, assuming the child path is
      * contained within the parent path, return the relative path from the
      * parent to the child. */
    public static String getRelativePath(File parent, File child) {
        def parentPath = parent.canonicalPath.split("[\\\\/]")
        def childPath = child.canonicalPath.split("[\\\\/]")

        /// If the parent path is longer it cannot contain the child path and
        /// we cannot construct a relative path without backtracking.
        if (parentPath.length > childPath.length) return ""

        /// Compare the parent and child path up until the end of the parent
        /// path.
        int b = 0
        while (b < parentPath.length && parentPath[b] == childPath[b] ) b++

        /// If we stopped before reaching the end of the parent path it must be
        /// that the paths do not match. The parent cannot contain the child and
        /// we cannot build a relative path without backtracking.
        if (b != parentPath.length) return ""
        return (['.'] + childPath[b..<childPath.length]).join('/') }


    public static Integer safeToInteger(def val) {
        if (val == null) return null
        try { return val as Integer }
        catch (NumberFormatException nfe) { return null } }
}
