package com.jdbernard.wdiwtlt.db

import java.lang.reflect.Modifier
import java.sql.Connection
import java.sql.PreparedStatement
import java.sql.ResultSet
import java.text.SimpleDateFormat
import java.util.regex.Pattern
import javax.persistence.Entity
import javax.sql.DataSource

import groovy.sql.Sql
import groovy.transform.CompileStatic
import groovy.transform.TailRecursive

import org.slf4j.Logger
import org.slf4j.LoggerFactory

import com.jdbernard.wdiwtlt.db.models.*

public class ORM {

    private DataSource dataSource
    private Sql sql

    private static UPPERCASE_PATTERN = Pattern.compile(/(.)(\p{javaUpperCase})/)
    private static Logger logger = LoggerFactory.getLogger(ORM)

    public ORM(DataSource dataSource) {
        this.dataSource = dataSource
        this.sql = new Sql(dataSource) }

    public void shutdown() { dataSource.shutdown() }

    /// ### Common
    public def getAll(Class modelClass) {
        def query = "SELECT * FROM " +
            pluralize(nameFromModel(modelClass.simpleName))
        logger.debug("Selecting models.\n\tSQL: {}", query)
        return sql.rows(query).collect { recordToModel(modelClass, it) } }

    public def getById(Class modelClass, int id) {
        def model = modelClass.instantiate()
        model.id = id
        return refresh(model) }

    public def refresh(def model) {
        def query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(" WHERE id = ?")
            .toString()

        logger.debug("Selecting model.\n\tSQL: {}\n\tPARAMS: {}", query, model.id)
        return updateModel(sql.firstRow(query, [id]), model) }

    public def getByName(Class modelClass, String name) {
        def query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(" WHERE name = ?")
            .toString()

        logger.debug("Selecting model.\n\tSQL: {}\n\tPARAMS: {}", query, name)
        return sql.rows(query, [name]).collect { recordToModel(it, modelClass) } }

    public def getBy(Class modelClass, List<String> columns,
    List<Object> values) {
        def query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(" WHERE ")
            .append(columns.collect { it + " = ?" }.join(' AND '))
            .toString()

        logger.debug("Selecting models.\n\tSQL: {}\n\tPARAMS: {}", query, values)
        return sql.rows(query, values)
            .collect { recordToModel(it, modelClass) } }

    public def getLike(Class modelClass, List<String> columns,
    List<Object> values) {
        values = values.collect { "%$it%".toString() }
        String query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(" WHERE ")
            .append(columns.collect { """"$it" LIKE ?"""}.join(' AND '))
            .toString()

        logger.debug("Selecting models.\n\tSQL: {}\n\tPARAMS: {}", query, values)
        return sql.rows(query, values)
            .collect { recordToModel(it, modelClass) } }

    public def save(def model) {
        if (model.id > 0) return update(model)
        else return create(model) }

    public def update(def model) {
        def setClauses = []
        def params = []

        getInstanceFields(model.class)
            .findAll { it.name != 'id' }
            .each { field ->
                setClauses << '"' + nameFromModel(field.name) + '"= ?'
                params << field.get(model) }

        def query = new StringBuilder()
            .append("UPDATE ")
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(" SET ")
            .append(setClauses.join(', '))
            .append(" WHERE id = ")
            .append(model.id)
            .toString()

        logger.debug("Updating model.\n\tSQL: {}\n\tPARAMS: {}", query, params)
        sql.executeUpdate(query, params)
        return model }

    public def create(def model) {
        def columns = []
        def params = []

        getInstanceFields(model.class)
            .findAll { it.name != 'id' }
            .each { field ->
                //if (field.class.getAnnotation(Entity)) // check to see if we
                                                        // have nested models
                columns << '"' + nameFromModel(field.name) + '"'
                params << field.get(model) }

        def query= new StringBuilder()
            .append("INSERT INTO ")
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(" (")
            .append(columns.join(', '))
            .append(") VALUES (")
            .append((1..columns.size()).collect { '?' }.join(', '))
            .append(")").toString()

        logger.debug("Creating model.\n\tSQL: {}\n\tPARAMS: {}", query, params)
        model.id = sql.executeInsert(query, params)[0][0]
        return model }

    public def delete(def model) {
        def query = new StringBuilder()
            .append("DELETE FROM ")
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append("WHERE id = ?")
            .toString()

        logger.debug("Deleting model.\n\tSQL: {}\n\tPARAMS: {}", query, model.id)
        sql.execute(query, [model.id])
        return sql.updateCount }

    public def associate(Class modelClass1, Class modelClass2 , Integer firstId, Integer secondId) {
        String linkTable = pluralize(nameFromModel(modelClass1.simpleName)) +
                     "_" + pluralize(nameFromModel(modelClass2.simpleName))
        String col1 = nameFromModel(modelClass1.simpleName) + "_id"
        String col2 = nameFromModel(modelClass2.simpleName) + "_id"

        withTransaction {
            def query = """\
                SELECT * FROM $linkTable
                WHERE "${col1}" = ? AND "${col2}" = ?"""
            def params = [firstId, secondId]

            // Look first for an existing association before creating one.
            logger.debug("Selecting association.\n\tSQL: {}\n\tPARAMS: {}",
                query, params)
            if (sql.firstRow(query, params)) { return 0 }
            else {
                query = """INSERT INTO $linkTable ("$col1", "$col2") VALUES (?, ?)"""
                logger.debug("Creating association.\n\tSQL: {}\n\tPARAMS: {}",
                    query, params)
                return sql.execute(query, params) } } }

    public def associate(def m1, def m2) {
        return associate(m1.class, m2.class, m1.id, m2.id) }

    /// ### Album-specific methods
    public Album getAlbumById(int id) { return getById(Album, id) }

    public Album[] getAlbumsByNameAndYear(String name, int year) {
        return getBy(Album, ['name', 'year'], [name, year]) }

    public List<Album> getAlbumsByName(String name) {
        return getByName(Album, name) }

    public Album getAlbumByNameAndArtistId(String name, int artistId) {
        def query = """\
            SELECT al.*
            FROM albums al JOIN
                 artists_albums aa ON
                    al.id = aa.album_id AND
                    aa.artist_id = ?
            WHERE al.name = ?"""
        def params = [artistId, name]
        logger.debug("Selecting an album.\n\tSQL: {}\n\tPARAMS: {}",
            query, params)
        return recordToModel(Album, sql.firstRow(query, params)) }

    public Album getAlbumByNameAndYearAndArtistId(String name, int year,
    int artistId) {
        def query = """\
            SELECT al.*
            FROM albums al JOIN
                 artists_albums aa ON
                 al.id = aa.album_id AND
                 aa.artist_id = ?
            WHERE al.name = ? AND al.year = ?"""
        def params = [artistId, name, year]
        logger.debug("Selecting an album. \n\tSQL: {}\n\tPARAMS: {}",
            query, params)
        return recordToModel(Album, sql.firstRow(query, params)) }

    public List<Album> getAlbumsByArtistId(int artistId) {
        def query = """\
            SELECT al.*
            FROM albums al JOIN
                 artists_albums aa ON
                    al.id = aa.album_id AND
                    aa.artist_id = ?"""

        logger.debug("Selecting albums.\n\tSQL: {}\n\tPARAMS: {}", query, artistId)
        return sql.rows(query, [artistId])
            .collect { recordToModel(Album, it) } }

    public List<Album> getAlbums() { return getAll(Album) }
    public List<Album> getAlbumsLikeName(String name) {
        return getLike(Album, ["name"], [name]) }

    public List<Album> removeEmptyAlbums() {
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    /// ### Artist-specific methods
    public Artist getArtistById(int id) { return getById(Artist, id) }
    public Artist getArtistByName(String name) {
        def artists = getByName(Artist, name)
        if (artists) return artists[0] }

    public List<Artist> getArtists() { return getAll(Artist) }
    public List<Artist> getArtistsByAlbumId(int albumId) {
        var query = """\
            SELECT ar.*
            FROM artists ar JOIN
                 artists_albums aa ON
                    ar.id = aa.artist_id AND
                    aa.album_id = ?"""
        logger.debug("Selecting artists.\n\tSQL: {}\n\tPARAMS: {}", query, artistId)
        return sql.rows(query, [albumId])
            .collect { recordToModel(Artist, it) } }

    public List<Artist> getArtistsLikeName(String name) {
        return getLike(Artist, ["name"], [name]) }

    public List<Artist> removeEmptyArtists() {
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    public def addAlbumArtist(int albumId, int artistId) {
        return associate(Artist, Album, artistId, albumId) }

    /// ### Bookmark-specific methods
    public Bookmark getBookmarkById(int id) { return getById(Bookmark, id) }

    public List<Bookmark> getBookmarkByName(String name) {
        return getByName(Bookmark, name) }

    public List<Bookmark> getBookmarks() { return getAll(Bookmark) }

    /// ### Image-specific methods
    public Image getImageById(int id) { return getById(Image, id) }

    /// ### MediaFile-specific methods
    public MediaFile getMediaFileById(int id) { return getById(MediaFile, id) }

    public List<MediaFile> getMediaFileByName(String name) {
        return getByName(MediaFile, name) }

    public List<MediaFile> getMediaFiles() { return getAll(MediaFile) }

    public List<MediaFile> getMediaFilesLikeName(String name) {
        return getLike(MediaFile, ["name"], [name]) }

    public MediaFile getMediaFileByFilePath(String filePath) {
        def files = getBy(MediaFile, ["file_path"], [filePath])
        return files ? files[0] : null }

    public def associateMediaFileWithAlbum(int mediaFileId, int albumId) {
        return associate(Album, MediaFile, albumId, mediaFileId) }

    public def associateMediaFileWithArtist(int mediaFileId, int artistId) {
        return associate(Album, MediaFile, artistId, mediaFileId) }

    public def incrementPlayCount(int mediaFileId) {
        def query = "UPDATE media_files SET play_count = play_count + 1 WHERE ID = ?"
        def params = [mediaFileId]

        logger.debug("Updating media file.\n\tSQL: {}\n\tPARAMS: {}", query, params)
        sql.executeUpdate(query, params)

        query = "SELECT play_count FROM media_files WHERE id = ?"
        logger.debug("Selecting media file play count.\n\tSQL: {}\n\tPARAMS: {}", query, params)
        return sql.firstRow(query, params)[0] }

    public MediaFile incrementPlayCount(MediaFile mf) {
        mf.playCount = incrementPlayCount(mf.id)
        return mf }

    /// ### Playlist-specific methods
    public Playlist getPlaylistById(int id) { return getById(Playlist, id) }

    public List<Playlist> getPlaylistByName(String name) {
        return getByName(Playlist, name) }

    public List<Playlist> getPlaylists() { return getAll(Playlist) }

    public Playlist addToPlaylist(int playlistId, int mediaFileId) {
        return withTransaction {
            associate(Playlist, MediaFile, p, mf)
            def p = getById(Playlist, playlistId)
            p.mediaFileCount++
            update(p)
            return refresh(p) } }

    public List<Playlist> removeEmptyPlaylists() {
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    /// ### Tag-specific methods
    public Tag getTagById(int id) { return getById(Tag, id) }
    public Tag getTagByName(String name) { return getByName(Tag, name)[0] }

    /// ### Utility functions
    public void withTransaction(Closure c) { 
        try { sql.execute("BEGIN TRANSACTION"); c() }
        finally { sql.execute("COMMIT") }
    }
    public static String nameToModel(String name) {
        def pts = name.toLowerCase().split('_')
        return pts.length == 1 ? pts[0] :
            pts[0] + pts[1..-1].collect { it.capitalize() }.join() }

    public static String nameFromModel(String name) {
        return UPPERCASE_PATTERN.matcher(name).
            replaceAll(/$1_$2/).toLowerCase() }

    public static String pluralize(String name) { return name + "s" }

    static def updateModel(def record, def model) {
        getInstanceFields(model.class).each { field ->
            field.set(model, record[nameFromModel(field.name)]) }
        return model }

    static def recordToModel(Class modelClass, def record) {
        if (record == null) return null

        def model = modelClass.newInstance()

        getInstanceFields(model.class).each { field ->
            field.set(model, record[nameFromModel(field.name)]) }

        return model }

    static def modelToRecord(def model) {
        if (model == null) return null

        def record = [:]

        getInstanceFields(model.class).each { field ->
            record[nameFromModel(field.name)] = field.get(model) }

        return record }

    static def getInstanceFields(Class modelClass) {
        return modelClass.fields.findAll { !Modifier.isStatic(it.modifiers) } }
}
